from pkgutil import iter_modules
from pathlib import Path
from importlib import import_module
import time
import traceback


def parse_modules(list_modules):
    modules = {}
    for module_name in list_modules:
        retries = 10
        while retries >= 0:
            try:
                addon = getattr(import_module(f"{__name__}.{module_name}"), 'Addon')
                addon.service = module_name
                modules[module_name] = addon
                retries = -1
                print(f"Successfully loaded addon: {module_name}")
            except Exception as e:
                print("----------------")
                print(f"Error with module: {module_name}")
                traceback.print_exc()
                time.sleep(2)
                retries -= 1
    return modules



#!/usr/bin/env python3

import os
import yaml
import time
import signal
import threading
import json
import traceback
import importlib.metadata
import platform
import argparse
import paho.mqtt.client as mqtt
from . import modules
from . import config

version = importlib.metadata.version(__package__ or __name__)


class GracefulKiller:
    kill_now = False

    def __init__(self):
        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)

    def exit_gracefully(self, signum, frame):
        print("LNXLink stopped by user")
        self.kill_now = True


class LNXlink():
    client = mqtt.Client()
    pref_topic = 'lnxlink'

    def __init__(self, config_path):
        print(f"LNXLink {version} started: {platform.python_version()}")
        self.config = self.read_config(config_path)

        modules_list = set(self.config['control'] + self.config['monitoring'])
        self.Addons = {}
        for service, addon in modules.parse_modules(modules_list).items():
            self.Addons[addon.service] = addon()

        self.setup_mqtt()

    def monitor_run(self):
        if self.config['monitoring'] is not None:
            for service in self.config['monitoring']:
                try:
                    addon = self.Addons[service]
                    subtopic = addon.name.lower().replace(' ', '/')
                    topic = f"{self.pref_topic}/{self.config['mqtt']['statsPrefix']}/{subtopic}"
                    pub_data = addon.getInfo()
                    # print(topic, pub_data, type(pub_data))
                    if type(pub_data) in [dict, list]:
                        pub_data = json.dumps(pub_data)
                    self.client.publish(
                        topic,
                        payload=pub_data,
                        retain=self.config['mqtt']['lwt']['retain']
                    )
                except Exception as e:
                    print(f"Can't load addon: {service}")
                    traceback.print_exc()

    def monitor_run_thread(self):
        self.monitor_run()

        self.monitor = threading.Timer(5.0, self.monitor_run_thread)
        self.monitor.start()

    def setup_mqtt(self):
        self.client = mqtt.Client()
        self.client.on_connect = self.on_connect
        self.client.on_message = self.on_message

        self.client.username_pw_set(self.config['mqtt']['auth']['user'], self.config['mqtt']['auth']['pass'])
        self.client.connect(self.config['mqtt']['server'], self.config['mqtt']['port'], 60)
        self.client.loop_start()

    def read_config(self, config_path):
        with open(config_path) as file:
            config = yaml.load(file, Loader=yaml.FullLoader)

        if 'prefix' in config['mqtt'] and 'clientId' in config['mqtt']:
            self.pref_topic = f"{config['mqtt']['prefix']}/{config['mqtt']['clientId']}"
        self.pref_topic = self.pref_topic.lower()

        config['monitoring'] = [x.lower().replace('-', '_') for x in config['monitoring']]
        config['control'] = [x.lower().replace('-', '_') for x in config['control']]
        return config

    def on_connect(self, client, userdata, flags, rc):
        print(f"Connected to MQTT with code {rc}")
        client.subscribe(f"{self.pref_topic}/commands/#")
        if self.config['mqtt']['lwt']['enabled']:
            self.client.publish(
                f"{self.pref_topic}/lwt",
                payload=self.config['mqtt']['lwt']['connectMsg'],
                qos=self.config['mqtt']['lwt']['qos'],
                retain=self.config['mqtt']['lwt']['retain']
            )
        if self.config['mqtt']['discovery']['enabled']:
            self.setup_discovery()

    def disconnect(self):
        print("Disconnected from MQTT.")
        if self.config['mqtt']['lwt']['enabled']:
            self.client.publish(
                f"{self.pref_topic}/lwt",
                payload=self.config['mqtt']['lwt']['disconnectMsg'],
                qos=self.config['mqtt']['lwt']['qos'],
                retain=self.config['mqtt']['lwt']['retain']
            )
        try:
            self.monitor.cancel()
        except Exception as e:
            pass
        self.client.disconnect()

    def on_message(self, client, userdata, msg):
        topic = msg.topic.replace(f"{self.pref_topic}/commands/", "")
        message = msg.payload
        print(f"Message received: {topic}")
        try:
            message = json.loads(message)
        except Exception as e:
            message = message.decode()
            print("String could not be converted to JSON")
            # traceback.print_exc()

        select_service = topic.split('/')
        control = self.Addons.get(select_service[0])
        if select_service[0] in self.config['control'] and control is not None:
            try:
                control.startControl(select_service, message)
                self.monitor_run()
            except Exception as e:
                traceback.print_exc()

    def setup_discovery(self):
        discovery_template = {
            "availability": {
                "topic": f"{self.pref_topic}/lwt",
                "payload_available": "ON",
                "payload_not_available": "OFF",
            },
            "device": {
                "identifiers": [self.config['mqtt']['clientId']],
                "name": self.config['mqtt']['clientId'],
                "model": self.config['mqtt']['prefix'],
                "manufacturer": f"LNXLink {version}"
            },
        }
        if self.config['monitoring'] is not None:
            for service in self.config['monitoring']:
                try:
                    addon = self.Addons[service]
                    subtopic = addon.name.lower().replace(' ', '/')
                    topic = f"{self.pref_topic}/{self.config['mqtt']['statsPrefix']}/{subtopic}"

                    discovery = discovery_template.copy()
                    discovery['name'] = addon.name.lower().replace(' ', '_')
                    discovery['unique_id'] = f"{self.config['mqtt']['clientId']}_{service}"
                    discovery['state_topic'] = topic
                    discovery['json_attributes_topic'] = topic
                    discovery['icon'] = addon.icon
                    discovery['unit_of_measurement'] = addon.unit
                    if hasattr(addon, 'device_class'):
                        # print(addon.device_class)
                        discovery['device_class'] = addon.device_class

                    if addon.unit == 'json':
                        discovery['unit_of_measurement'] = ""
                        discovery['value_template'] = "{{ value_json.status }}"
                        discovery['json_attributes_template'] = "{{ value_json | tojson }}"

                    sensor_type = getattr(addon, 'sensor_type', 'sensor')
                    if sensor_type in ['binary_sensor', 'sensor']:
                        self.client.publish(
                            f"homeassistant/{sensor_type}/lnxlink/{discovery['unique_id']}/config",
                            payload=json.dumps(discovery),
                            retain=self.config['mqtt']['lwt']['retain'])
                except Exception as e:
                    traceback.print_exc()
        if self.config['monitoring'] is not None:
            for service in self.config['control']:
                addon = self.Addons.get(service)
                if addon is not None and hasattr(addon, 'exposedControls'):
                    for control_name, options in addon.exposedControls().items():
                        try:
                            discovery = discovery_template.copy()
                            discovery['name'] = control_name.lower().replace(' ', '_')
                            discovery['unique_id'] = f"{self.config['mqtt']['clientId']}_{control_name}"
                            discovery['icon'] = options.get('icon', '')

                            if options['type'] == 'button':
                                discovery['state_topic'] = f"{self.pref_topic}/lwt"
                                discovery["command_topic"] = f"{self.pref_topic}/commands/{service}/{control_name}/"
                            elif options['type'] == 'switch':
                                discovery["state_topic"] = f"{self.pref_topic}/{self.config['mqtt']['statsPrefix']}/{subtopic}"
                                discovery["command_topic"] = f"{self.pref_topic}/commands/{service}/{control_name}/"
                                discovery["payload_off"] = "OFF"
                                discovery["payload_on"] = "ON"
                            elif options['type'] == 'number':
                                continue
                            else:
                                continue
                            self.client.publish(
                                f"homeassistant/{options['type']}/lnxlink/{discovery['unique_id']}/config",
                                payload=json.dumps(discovery),
                                retain=self.config['mqtt']['lwt']['retain'])
                        except Exception as e:
                            traceback.print_exc()


def main():
    parser = argparse.ArgumentParser(
        prog="LNX Link",
        description="Send system information to MQTT broker")
    parser.add_argument(
        "-c", "--config",
        help="Configuration file",
        default="/etc/config.yaml",
        required=True)
    args = parser.parse_args()

    config_file = os.path.abspath(args.config)
    config.setup_config(config_file)
    config.setup_systemd(config_file)
    lnxlink = LNXlink(config_file)
    lnxlink.monitor_run_thread()

    killer = GracefulKiller()
    while not killer.kill_now:
        time.sleep(1)
    lnxlink.disconnect()


if __name__ == '__main__':
    main()



import psutil
from datetime import datetime


class Addon():
    name = 'Network Upload'
    icon = 'mdi:access-point-network'
    unit = 'Mbit/s'

    def __init__(self):
        self.timeOld = datetime.now()
        self.sentOld = psutil.net_io_counters().bytes_sent

    def getInfo(self):
        """ Returns Mbps"""
        timeNew = datetime.now()
        sentNew = psutil.net_io_counters().bytes_sent

        timeDiff = (timeNew - self.timeOld).total_seconds()
        sentDiff = sentNew - self.sentOld

        self.timeOld = timeNew
        self.sentOld = sentNew

        sentsSpeed = round(sentDiff * 8 / timeDiff / 1024 / 1024, 2)

        return sentsSpeed



import subprocess

class Addon():
    name = 'Shutdown'
    icon = 'mdi:power'
    unit = 'button'

    def startControl(self, topic, data):
        subprocess.call(["shutdown", "now"])

    def exposedControls(self):
        return {
            "shutdown": {
                "type": "button",
                "icon": "mdi:power",
            }
        }



from idle_time import IdleMonitor


class Addon():
    name = 'Idle'
    icon = 'mdi:timer-sand'
    unit = 'min'

    def getInfo(self):
        monitor = IdleMonitor.get_monitor()
        idle_sec = monitor.get_idle_time()
        idle_min = int(idle_sec / 60)
        return idle_min



import subprocess


class Addon():
    name = 'Screen OnOff'
    icon = 'mdi:monitor'
    unit = None
    sensor_type = 'switch'

    def exposedControls(self):
        return {
            "Screen_OnOff": {
                "type": "switch",
                "icon": "mdi:monitor",
            }
        }

    def getInfo(self):
        stdout = subprocess.run(
            "xset q | grep -i 'monitor is'",
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE).stdout.decode("UTF-8")
        results = stdout.lower()

        status = results.strip().replace('monitor is', '').strip().upper()
        return status

    def startControl(self, topic, data):
        if data.lower() == 'off':
            subprocess.call('xset dpms force off', shell=True)
        elif data.lower() == 'on':
            subprocess.call('xset dpms force on', shell=True)



import notify2
import requests

class Addon():
    name = 'Notify OSD'
    icon = None
    unit = None

    def startControl(self, topic, data):
        iconUrl = data.get('iconUrl', '')
        icon = 'notification-message-im'
        if 'http' in iconUrl:
            iconExt = iconUrl.split('.')[-1]
            icon = f"/tmp/lnxlink_notification.{iconExt}"
            img_data = requests.get(iconUrl).content
            with open(icon, 'wb') as handler:
                handler.write(img_data)
        elif iconUrl != '':
            icon = iconUrl

        # notify2
        notify2.init("Test")
        notify2.Notification(
            data['title'],
            data['message'],
            icon
        ).show()

        print(topic, data)




import psutil
from datetime import datetime


class Addon():
    name = 'Network Download'
    icon = 'mdi:access-point-network'
    unit = 'Mbit/s'

    def __init__(self):
        self.timeOld = datetime.now()
        self.recvOld = psutil.net_io_counters().bytes_recv

    def getInfo(self):
        """ Returns Mbps"""
        timeNew = datetime.now()
        recvNew = psutil.net_io_counters().bytes_recv

        timeDiff = (timeNew - self.timeOld).total_seconds()
        recvDiff = recvNew - self.recvOld

        self.timeOld = timeNew
        self.recvOld = recvNew

        recvSpeed = round(recvDiff * 8 / timeDiff / 1024 / 1024, 2)

        return recvSpeed



#!/usr/bin/env python3

import yaml
import sys
import requests
import os
import subprocess
import shutil
from pathlib import Path


github_repository = "bkbilly/lnxlink/master"


def setup_config(config_path):
    if not os.path.exists(config_path):
        print("Config file not found.")
        url = f"https://raw.githubusercontent.com/{github_repository}/config_temp.yaml"
        r = requests.get(url)
        try:
            Path(config_path).parent.mkdir(parents=True, exist_ok=True)
            with open(config_path, 'wb') as config:
                config.write(r.content)
            print(f"Created new template: {config_path}")
        except IOError:
            print("Permision denied")
            return False
        userprompt_config(config_path)
    return True


def query_true_false(question, default="false"):
    valid = {"true": True, "t": True, "yes": True, True: True, "false": False, "f": False, "no": False, False:False}

    if default == True:
        prompt = "(True/False) [True]"
    elif default == False or default is None:
        prompt = "(True/False) [False]"
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        choice = input(f" {question} {prompt}: ").lower()
        if default is not None and choice == "":
            return valid[default]
        elif choice in valid:
            return valid[choice]
        else:
            print("Please respond with 'true' or 'false' (or 't' or 'f').")


def userprompt_config(config_path):
    with open(config_path) as f:
        y = yaml.safe_load(f)

        # Print config
        print('This will update the MQTT credentials and topics, these are the default topics:')
        print(" MQTT Topic prefix for for monitoring: {}/{}/{}/...".format(
            y['mqtt']['prefix'],
            y['mqtt']['clientId'],
            y['mqtt']['statsPrefix']
        ))
        print(" MQTT Topic prefix for for commands: {}/{}/commands/...".format(
            y['mqtt']['prefix'],
            y['mqtt']['clientId']
        ))

        print("\nLeave empty for default")

        # Change default values
        y['mqtt']['discovery']['enabled'] = query_true_false("Enable MQTT automatic discovery", y['mqtt']['discovery']['enabled'])
        y['mqtt']['server'] = input(f" MQTT server [{y['mqtt']['server']}]: ") or y['mqtt']['server']
        y['mqtt']['port'] = input(f" MQTT port [{y['mqtt']['port']}]: ") or y['mqtt']['port']
        y['mqtt']['port'] = int(y['mqtt']['port'])
        y['mqtt']['auth']['user'] = input(f" MQTT username [{y['mqtt']['auth']['user']}]: ") or y['mqtt']['auth']['user']
        y['mqtt']['auth']['pass'] = input(f" MQTT password [{y['mqtt']['auth']['pass']}]: ") or y['mqtt']['auth']['pass']

        y['mqtt']['prefix'] = input(f" Change prefix [{y['mqtt']['prefix']}]: ") or y['mqtt']['prefix']
        y['mqtt']['clientId'] = input(f" Change clientId [{y['mqtt']['clientId']}]: ") or y['mqtt']['clientId']
        y['mqtt']['statsPrefix'] = input(f" Change statsPrefix [{y['mqtt']['statsPrefix']}]: ") or y['mqtt']['statsPrefix']


    with open(config_path, 'w') as fw:
        fw.write(yaml.dump(y, default_flow_style=False, sort_keys=False))

    print('\nAll changes have been saved.')
    print(" MQTT Topic prefix for for monitoring: {}/{}/{}/...".format(
        y['mqtt']['prefix'],
        y['mqtt']['clientId'],
        y['mqtt']['statsPrefix']
    ))
    print(" MQTT Topic prefix for for commands: {}/{}/commands/...".format(
        y['mqtt']['prefix'],
        y['mqtt']['clientId']
    ))


def get_service_user():
    installed_as = 0
    for num, cmd_user in enumerate(["--user", ""], start=1):
        cmd = f"systemctl {cmd_user} is-active lnxlink.service"
        stdout = subprocess.run(
            cmd,
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE).stdout.decode("UTF-8")
        result = stdout.strip()
        if result in ['active', 'failed']:
            _, _, _, service_location = get_service_vars(num)
            if os.path.exists(f"{service_location}/lnxlink.service"):
                installed_as = num
    return installed_as


def get_service_vars(user_service):
    if user_service == True or user_service == 1:
        sudo = ""
        cmd_user = "--user"
        service_url = f"https://raw.githubusercontent.com/{github_repository}/autostart/lnxlink_user.service"
        service_location = f"{os.path.expanduser('~')}/.config/systemd/user"
    else:
        sudo = "sudo"
        cmd_user = ""
        service_url = f"https://raw.githubusercontent.com/{github_repository}/autostart/lnxlink_headless.service"
        service_location = "/etc/systemd/system"
    return sudo, cmd_user, service_url, service_location


def setup_systemd(config_path):
    # Check how systemd is installed
    installed_as = get_service_user()

    if installed_as == 0:
        # Service not found or not running
        print("SystemD service not found or it's not running...")
        user_service = query_true_false("Install as a user service?", True)
        sudo, cmd_user, service_url, service_location = get_service_vars(user_service)

        # Install on SystemD
        Path(service_location).mkdir(parents=True, exist_ok=True)
        r = requests.get(service_url)
        with open(f"{service_location}/lnxlink.service", 'wb') as config:
            exec_cmd = f"{shutil.which('lnxlink')} -c {config_path}"
            content = r.content.replace(b"{exec_cmd}", exec_cmd.encode())
            config.write(content)

        cmd = f"{sudo} chmod +x {service_location}/lnxlink.service"
        subprocess.call(cmd, shell=True)
        cmd = f"{sudo} systemctl {cmd_user} enable lnxlink.service"
        subprocess.call(cmd, shell=True)
        cmd = f"{sudo} systemctl {cmd_user} daemon-reload"
        subprocess.call(cmd, shell=True)


if __name__ == '__main__':
    # config_path = sys.argv[1]
    # download_template(config_path)
    setup_systemd("lnxlink.yaml")



import glob


class Addon():
    name = 'Microphone used'
    icon = 'mdi:microphone'
    sensor_type = 'binary_sensor'
    unit = ''

    def getInfo(self):
        mics = glob.glob('/proc/asound/**/*c/sub*/status', recursive=True)
        for mic in mics:
            with open(mic) as mic_content:
                mic_status = mic_content.read().strip().lower()
                if mic_status != 'closed':
                    return "ON"
        return "OFF"



# from pynput.keyboard import Key, Controller
from pynput import keyboard

class Addon():
    name = 'Send Keys'
    icon = None
    unit = None

    def startControl(self, topic, data):
        if type(data) == str:
            self.__presskeys(data)
        else:
            for keys in data:
                self.__presskeys(keys)

    def __presskeys(self, keys):
        contr = keyboard.Controller()
        hotkeys = keyboard.HotKey.parse(keys)
        for hotkey in hotkeys:
            contr.press(hotkey)
        hotkeys.reverse()
        for hotkey in hotkeys:
            contr.release(hotkey)


if __name__ == '__main__':
    myaddon = Addon()
    # myaddon.startControl('send-keys', ['<alt>+<F4>', '<alt>+<ctrl>+<delete>', '<ctrl>+C'])
    # myaddon.startControl('send-keys', ['<shift>+C', 'c'])



import subprocess
import re


class Addon():
    name = 'Bluetooth Battery'
    icon = 'mdi:battery'
    device_class = 'battery'
    unit = '%'

    def getInfo(self):
        stdout = subprocess.run(
            'upower --dump | grep percentage',
            shell=True,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE).stdout.decode("UTF-8")
        results = stdout.strip().split()

        percentage = -1
        if len(results) >= 2:
            if re.match(r'\d+%', results[1]):
                percentage = int(results[1].replace("%", ""))

        return percentage
